@@ -27,8 +27,8 @@
 #
 #   Sequel.pg_row(array)
 #
-# If you have loaded the {core_extensions extension}[link:files/doc/core_extensions_rdoc.html]),
-# or you have loaded the {core_refinements extension}[link:files/doc/core_refinements_rdoc.html])
+# If you have loaded the {core_extensions extension}[rdoc-ref:doc/core_extensions.rdoc],
+# or you have loaded the core_refinements extension
 # and have activated refinements for the file, you can also use Array#pg_row:
 #
 #   array.pg_row
@@ -66,6 +66,21 @@
 #
 #   DB[:table].insert(:address=>DB.row_type(:address, :street=>'123 Sesame St.', :city=>'Some City', :zip=>'12345'))
 #
+# Note that registering row types without providing an explicit :converter option
+# creates anonymous classes.  This results in ruby being unable to Marshal such
+# objects.  You can work around this by assigning the anonymous class to a constant.
+# To get a list of such anonymous classes, you can use the following code:
+#
+#   DB.conversion_procs.select{|k,v| v.is_a?(Sequel::Postgres::PGRow::Parser) && \
+#     v.converter && (v.converter.name.nil? || v.converter.name == '') }.map{|k,v| v}
+# 
+# If you are not using the native postgres adapter and are using composite types
+# types as model column values you probably should use the
+# pg_typecast_on_load plugin if the column values are returned as a string.
+#
+# See the {schema modification guide}[rdoc-ref:doc/schema_modification.rdoc]
+# for details on using row type columns in CREATE/ALTER TABLE statements.
+#
 # This extension requires both the strscan and delegate libraries.
 
 require 'delegate'
@@ -98,7 +113,7 @@ module Sequel
         # This is done so that instances of this subclass are
         # automatically casted to the database type when literalizing.
         def self.subclass(db_type)
-          sc = Class.new(self) do
+          Class.new(self) do
             @db_type = db_type
           end
         end
@@ -146,7 +161,7 @@ module Sequel
         # Create a new subclass of this class with the given database
         # type and columns.
         def self.subclass(db_type, columns)
-          sc = Class.new(self) do
+          Class.new(self) do
             @db_type = db_type
             @columns = columns
           end
@@ -196,6 +211,8 @@ module Sequel
         end
       end
 
+      ROW_TYPE_CLASSES = [HashRow, ArrayRow]
+
       # This parser-like class splits the PostgreSQL
       # row-valued/composite type output string format
       # into an array of strings.  Note this class makes
@@ -276,7 +293,7 @@ module Sequel
 
         # Sets each of the parser's attributes, using options with
         # the same name (e.g. :columns sets the columns attribute).
-        def initialize(h={})
+        def initialize(h=OPTS)
           @columns = h[:columns]
           @column_converters = h[:column_converters]
           @column_oids = h[:column_oids]
@@ -377,6 +394,7 @@ module Sequel
             @row_types = {}
             @row_schema_types = {}
             extend(@row_type_method_module = Module.new)
+            copy_conversion_procs([2249, 2287])
           end
         end
 
@@ -402,7 +420,7 @@ module Sequel
         #
         # :converter :: Use a custom converter for the parser.
         # :typecaster :: Use a custom typecaster for the parser.
-        def register_row_type(db_type, opts={})
+        def register_row_type(db_type, opts=OPTS)
           procs = @conversion_procs
           rel_oid = nil
           array_oid = nil
@@ -431,11 +449,12 @@ module Sequel
 
           # Get column names and oids for each of the members of the composite type.
           res = from(:pg_attribute).
+            join(:pg_type, :oid=>:atttypid).
             where(:attrelid=>rel_oid).
             where{attnum > 0}.
             exclude(:attisdropped).
             order(:attnum).
-            select_map([:attname, :atttypid])
+            select_map([:attname, Sequel.case({0=>:atttypid}, :pg_type__typbasetype, :pg_type__typbasetype).as(:atttypid)])
           if res.empty?
             raise Error, "no columns for row type #{db_type.inspect} in database"
           end
@@ -471,6 +490,7 @@ module Sequel
 
           @row_types[db_type] = opts.merge(:parser=>parser)
           @row_schema_types[schema_type_string] = schema_type_symbol 
+          @schema_type_classes[schema_type_symbol] = ROW_TYPE_CLASSES
           @row_type_method_module.class_eval do
             meth = :"typecast_value_#{schema_type_symbol}"
             define_method(meth) do |v|
@@ -479,6 +499,7 @@ module Sequel
             private meth
           end
 
+          conversion_procs_updated
           nil
         end
 
@@ -523,15 +544,6 @@ module Sequel
           end
         end
 
-        # Make the column type detection handle registered row types.
-        def schema_column_type(db_type)
-          if type = @row_schema_types[db_type]
-            type
-          else
-            super
-          end
-        end
-
         private
 
         # Format composite types used in bound variable arrays.
@@ -546,6 +558,15 @@ module Sequel
             super
           end
         end
+
+        # Make the column type detection handle registered row types.
+        def schema_column_type(db_type)
+          if type = @row_schema_types[db_type]
+            type
+          else
+            super
+          end
+        end
       end
     end
 
@@ -572,6 +593,7 @@ module Sequel
   Database.register_extension(:pg_row, Postgres::PGRow::DatabaseMethods)
 end
 
+# :nocov:
 if Sequel.core_extensions?
   class Array
     # Wraps the receiver in an anonymous Sequel::Postgres::PGRow::ArrayRow instance.
@@ -590,3 +612,4 @@ if defined?(Sequel::CoreRefinements)
     end
   end
 end
+# :nocov:
